/* -*-c++-*- source - Open Producer - Copyright (C) 2002 Don Burns
 * Distributed under the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE (LGPL)
 * as published by the Free Software Foundation.
 */
#include <string>
#include <map>
#include <Producer/CameraGroup>

using namespace Producer; 

void CameraGroup::_initVariables()
{
    _stack_size  = 0;
    _frame_count = 0;
    _sync_count  = 0;
    _block_on_vsync = false;
    char *ptr = getenv( "PRODUCER_CAMERA_BLOCK_ON_VSYNC" );
    if( ptr )
        _block_on_vsync = true;
    //_lens = new Camera::Lens;
    _realized    = false;
    _instrumented = false;
    _thread_model = SingleThreaded;
}

void CameraGroup::_initLens()
{
    if( !_lens.valid() )
    if( _cfg->getNumberOfCameras() > 0 )
        _lens = _cfg->getCamera(0)->getLens();

    for( unsigned int i = 0; i < _cfg->getNumberOfCameras(); i++ )
    {
        Camera *cam = _cfg->getCamera(i);
        cam->setLens( _lens.get() );
    }
}

CameraGroup::CameraGroup( CameraConfig *cfg )
{
    _initVariables();
    _cfg = cfg;
    _initLens();
}

CameraGroup::CameraGroup()
{
    _initVariables();
    _cfg = new CameraConfig;
    _cfg->defaultConfig();
    _initLens();
}

CameraGroup::CameraGroup( const std::string& configFile )
{
    _initVariables();
    _cfg = new CameraConfig;
    if( configFile.empty() )
        _cfg->defaultConfig();
    else
        _cfg->parseFile( configFile.c_str() );
    _initLens();
}


void CameraGroup::setStackSize( size_t size )
{
    if( _realized )
    {
        std::cerr << "CameraGroup::setStackSize(): Camera Group is already realized.\n"
                     "SetStackSize may only be called before CameraGroup::realize();\n";
        return ;
    }
    _stack_size = size;
}


bool CameraGroup::realize( ThreadingModel thread_model )
{
    if( _realized ) return _realized;
    _thread_model = thread_model;
    return realize();
}

bool CameraGroup::realize()
{
    if( _realized ) return _realized;

    _frameStats._frameTimeStampSets.resize(_cfg->getNumberOfCameras());
    _initTime = _timer.tick();


    if( _thread_model == SingleThreaded )
    {
        for( unsigned int i = 0; i < _cfg->getNumberOfCameras(); i++ )
        {
            RenderSurface *rs = _cfg->getCamera(i)->getRenderSurface();
#ifndef WIN32
            rs->useConfigEventThread(false);
#endif
            rs->realize();
            _cfg->getCamera(i)->setBlockOnVsync(_block_on_vsync);
        }
        _realized = true;
    }
    else if( _thread_model == ThreadPerCamera )
    {
#ifdef _X11_IMPLEMENTATION
        RenderSurface::InitThreads();
#endif
        int nthreads = _cfg->getNumberOfCameras();

        _sync_barrier  = new Barrier( nthreads+1 );
        _frame_barrier = new Barrier(nthreads+1);

        for( int i = 0; i < nthreads; i++ )
        {
            Camera *cam = _cfg->getCamera(i);
            cam->setSyncBarrier( _sync_barrier );
            cam->setFrameBarrier( _frame_barrier );
            if( _stack_size != 0 )
                cam->setStackSize( _stack_size );
            cam->setInitTime( _initTime );
            cam->setBlockOnVsync(_block_on_vsync);
            cam->startThread();
        }
        _realized = true;
    }
    else
    {
        std::cout << "CameraGroup::realize() : Threading model unsupported\n";
        return _realized ;
    }
    waitForRealize();
    return _realized;
}

void CameraGroup::setViewByLookat(  float eyex, float eyey, float eyez,
                                    float centerx, float centery, float centerz,
                                    float upx, float upy, float upz )
{
    setViewByLookat( Vec3(eyex, eyey, eyez),
                     Vec3(centerx, centery, centerz ),
                     Vec3(upx, upy, upz) );
}

void CameraGroup::setViewByLookat(  const Vec3& eye,
                                    const Vec3& center,
                                    const Vec3& up )
{
    Matrix m;
    m.makeLookAt(eye,center,up);
    setViewByMatrix( m );
}

void CameraGroup::setViewByMatrix(  const Producer::Matrix &mat )
{
    for( unsigned int i = 0; i < _cfg->getNumberOfCameras(); i++ )
        _cfg->getCamera(i)->setViewByMatrix( mat );
}

void CameraGroup::setInstrumentationMode( bool mode )
{
    _instrumented = mode;
    for( unsigned int i = 0 ; i < _cfg->getNumberOfCameras(); i++ )
        _cfg->getCamera(i)->setInstrumentationMode(_instrumented);
}


void CameraGroup::frame()
{
    if( _instrumented )
        _frameInstrumented();
    else
        _frame();
}

void CameraGroup::_frame()
{
    if( !_realized )
    {
        std::cerr << "CameraGroup::frame() : Please call realize() first\n";
        return;
    }
    if( _thread_model == SingleThreaded )
    {
        std::map <RenderSurface *, Camera *> u;
        for( unsigned int i = 0; i < _cfg->getNumberOfCameras(); i++ )
        {
            u.insert( std::pair <RenderSurface *, Camera *>(
                                    _cfg->getCamera(i)->getRenderSurface(),
                                    _cfg->getCamera(i) ));
            _cfg->getCamera(i)->frame(false);
        }

        std::map <RenderSurface *, Camera *>::iterator p;
        for( p = u.begin(); p != u.end(); p++ )
            p->second->advance();
    }
    else if( _thread_model == ThreadPerCamera )
    {
        if( _sync_count == _frame_count ) 
            sync();
        _frame_barrier->block();
    }
    else
    {
        std::cout << "CameraGroup::frame() : Threading model unsupported\n";
    }
    _frame_count++;
}

void CameraGroup::_frameInstrumented()
{
    if( !_realized )
    {
        std::cerr << "CameraGroup::frame() : Please call realize() first\n";
        return;
    }

    if( _sync_count == _frame_count ) 
        _syncInstrumented(false);

    if( _thread_model == SingleThreaded )
    {
        _endOfUpdate = _timer.tick();
        _startOfFrame = _timer.tick(); 
        std::map <RenderSurface *, Camera *> u;
        for( unsigned int i = 0; i < _cfg->getNumberOfCameras(); i++ ) 
        {
            u.insert( std::pair <RenderSurface *, Camera *>(
                                    _cfg->getCamera(i)->getRenderSurface(),
                                    _cfg->getCamera(i) ));
            _cfg->getCamera(i)->frame(false);
        }
        _startOfUpdate = _timer.tick();

        std::map <RenderSurface *, Camera *>::iterator p;
        for( p = u.begin(); p != u.end(); p++ )
            p->second->advance();
    }
    else if( _thread_model == ThreadPerCamera )
    {
        _endOfUpdate =  _timer.tick();
        _startOfFrame = _timer.tick();
        _frame_barrier->block();
    }
    else
    {
        std::cout << "CameraGroup::frame() : Threading model unsupported\n";
    }

    _frame_count++;
}

void CameraGroup::sync()
{
    if( _instrumented )
        _syncInstrumented(true);
    else
        _sync();
}

void CameraGroup::_sync()
{
    if( _sync_count > _frame_count ) return;

    if( _thread_model == SingleThreaded )
    {
    }
    else if( _thread_model == ThreadPerCamera )
    {
        _sync_barrier->block();
    }
    else
    {
        std::cout << "CameraGroup::frame() : Threading model unsupported\n";
    }
    _sync_count = _frame_count + 1;
}

void CameraGroup::_updateStats()
{
    _frameStats._frameNumber = _frame_count;
    _frameStats._startOfUpdate = _timer.delta_s(_initTime, _startOfUpdate );
    _frameStats._endOfUpdate = _timer.delta_s(_initTime, _endOfUpdate );
    _frameStats._startOfFrame = _timer.delta_s( _initTime, _startOfFrame );
    for( unsigned int i = 0; i < _cfg->getNumberOfCameras(); i++ )
    {
        _frameStats._frameTimeStampSets[i] = _cfg->getCamera(i)->getFrameStats();
        _frameStats._frameTimeStampSets[i].setFrameNumber(_frameStats._frameNumber);
    }

    if( _statsHandler.valid() )
        (*_statsHandler)(*this);
}

void CameraGroup::_syncInstrumented( bool markStartOfUpdate )
{
    if( _sync_count > _frame_count ) return;

    if( _thread_model == ThreadPerCamera )
        _sync_barrier->block();

    _updateStats();
    if( markStartOfUpdate )
        _startOfUpdate = _timer.tick();
    _sync_count = _frame_count + 1;
}


void CameraGroup::advance()
{
    /*
    for( unsigned int i = 0; i < _cfg->getNumberOfRenderSurfaces(); i++ )
    {
        _cfg->getRenderSurface(i)->makeCurrent();
        _cfg->getRenderSurface(i)->swapBuffers();
    }
    */

    for( unsigned int i = 0; i < _cfg->getNumberOfCameras(); i++ )
    {
    _cfg->getCamera(i)->getRenderSurface()->makeCurrent();
    _cfg->getCamera(i)->getRenderSurface()->swapBuffers();
    }
}

Camera *CameraGroup::getCamera(int i) 
{ 
    return _cfg->getCamera(i); 
}

const Camera *CameraGroup::getCamera(int i)  const
{ 
    return _cfg->getCamera(i); 
}

unsigned int CameraGroup::getNumberOfCameras() const
{
    return _cfg->getNumberOfCameras();
}

CameraConfig *CameraGroup::getCameraConfig()
{
    return _cfg;
}

const CameraConfig *CameraGroup::getCameraConfig() const
{
    return _cfg;
}

void CameraGroup::setSceneHandler(Camera::SceneHandler *sh )
{
    for( unsigned int i = 0; i <  _cfg->getNumberOfCameras(); i++ )
    _cfg->getCamera(i)->setSceneHandler(sh);    
}


#if 0
bool CameraGroup::convertLensToOrtho( float d) 
{ 
    bool retval = true;

    for( unsigned int i = 0; i <  _cfg->getNumberOfCameras(); i++ )
        if( _cfg->getCamera(i)->convertLensToOrtho(d) == false )
        retval = false;

    return retval;
}

bool CameraGroup::convertLensToPerspective( float d )
{
    bool retval = true;

    for( unsigned int i = 0; i <  _cfg->getNumberOfCameras(); i++ )
        if( _cfg->getCamera(i)->convertLensToPerspective(d) == false )
        retval = false;

    return retval;
}
#endif


bool CameraGroup::waitForRealize()
{
    bool retval = true;
    for( unsigned int i = 0; i < _cfg->getNumberOfCameras(); i++ )
    {
        if( _cfg->getCamera(i)->getRenderSurface()->waitForRealize() == false )
            retval = false;
    }
    return retval;
}

bool CameraGroup::validForRendering() const
{
    bool retval = true;
    for( unsigned int i = 0; i < _cfg->getNumberOfCameras(); i++ )
    {
        if( _cfg->getCamera(i)->getRenderSurface()->isRealized() == false )
        {
            retval = false;
            break;
        }
    }
    return retval;
}
