/* -*-c++-*- source - Open Producer - Copyright (C) 2002 Don Burns
 * Distributed under the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE (LGPL)
 * as published by the Free Software Foundation.
 */
#include <Producer/RenderSurface>
#include <Producer/Trackball>
#include <Producer/KeyboardMouse>
#include <Producer/Window3D>

using namespace Producer;
using namespace std;


class Window3DKbdMouseCallback : public KeyboardMouseCallback
{
    public:
        Window3DKbdMouseCallback() : KeyboardMouseCallback()
        {
            _tb = NULL;
            _button_state = 0;
            _clientKeyCallback = 0L;
            _done = false;
        }

        virtual void buttonPress( float mx , float my, unsigned int button) 
        {
            _mx = mx;
            _my = my;
            _button_state |= (1<<(button-1));

            if( _tb != NULL ) 
                _tb->input(_mx, _my, _button_state );
        }

        virtual void buttonRelease( float mx, float my, unsigned int button ) 
        {
            _mx = mx;
            _my = my;
            _button_state &= ~(1<<(button-1));

            if( _tb != NULL ) 
                _tb->input(_mx, _my, _button_state );
        }

        virtual void mouseMotion( float mx, float my) 
        {
            _mx = mx;
            _my = my;
        }

        void setTrackball( Trackball *tb ) { _tb = tb; }
        void setClientKeyboardCallback( Window3D::KeyboardCallback *callback )
        {
            _clientKeyCallback = callback;
        }

        float mx() { return _mx; }
        float my() { return _my; }
        unsigned int mbutton() { return _button_state; }

        virtual void specialKeyPress( KeyCharacter key )
        {
            switch( key )
            {
                case Producer::KeyChar_Escape:
                    _done = true;
                    break;
            }
            if( _clientKeyCallback != NULL )
                (*_clientKeyCallback)( key );
        }

        virtual void keyPress( KeyCharacter key )
        {
            switch( key )
            {
                case ' ':
                    _tb->reset();
                    break;

            }
            if( _clientKeyCallback != NULL )
                (*_clientKeyCallback)( key );
        }

        void shutdown() { _done = true; }

        bool done() { return _done; }

    private :
        Trackball *_tb;
        float _mx;
        float _my;
        bool _done;
        unsigned int _button_state;
        Window3D::KeyboardCallback *_clientKeyCallback;
};

class Window3D::Implementation {
    public :
        Implementation() 
        {
            _trackball = NULL;
            //_kbdMouse = NULL;
            _kbdMouse = new KeyboardMouse(&_rs);
        }

        RenderSurface &renderSurface() { return _rs; }
        Trackball *trackball() { return _trackball; }
        KeyboardMouse *kbdMouse() { return _kbdMouse; }
        Window3DKbdMouseCallback &kbdMouseCallback() { return kmcb; }
        void enableTrackball()
        {
            if( _kbdMouse == NULL ) 
                _kbdMouse = new KeyboardMouse(&_rs);
            if( _trackball != NULL ) return;
                _trackball = new Trackball;
            _trackball->setOrientation( Trackball::Y_UP );
            kmcb.setTrackball( _trackball );
        }

    void disableTrackball()
    {
        if( _trackball == NULL ) return;
        delete _trackball;
        _trackball = 0L;
    }

    void setKeyboardCallback( KeyboardCallback *callback )
    {
        if( _kbdMouse == NULL ) 
            _kbdMouse = new KeyboardMouse(&_rs);
        kmcb.setClientKeyboardCallback( callback );
    }

    void update()
    {
        if( _kbdMouse == NULL ) return;
            _kbdMouse->update(kmcb);

        if( _trackball == NULL ) return;
        _trackball->input( kmcb.mx(), kmcb.my(), kmcb.mbutton() );
    }

    bool done() { return kmcb.done(); }

    private :
        RenderSurface _rs;
        Trackball *_trackball;
        KeyboardMouse *_kbdMouse;
        Window3DKbdMouseCallback kmcb;
};

Window3D::Window3D(
        const std::string name,
        int posx,
        int posy,
        unsigned int width,
        unsigned int height,
        unsigned int parent )
{
    string _name;
    if( name == "" )
    _name = "Window3D"; 
    else
    _name = name;

    imp = new Implementation;
    imp->renderSurface().setWindowName( name );
    imp->renderSurface().setParentWindow( Window(parent) );

    if( width != (unsigned int)-1 || height != (unsigned int)-1 )
        imp->renderSurface().setWindowRectangle( posx, posy, width, height );

    imp->renderSurface().realize();
}

unsigned int Window3D::width( void ) const
{
    return imp->renderSurface().getWindowWidth();
}

unsigned int Window3D::height( void ) const
{
    return imp->renderSurface().getWindowHeight();
}

void Window3D::getDimensions( unsigned int &width, unsigned int &height )
{
    int  x, y;
    unsigned int w, h;
    imp->renderSurface().getWindowRectangle( x, y,w, h );
    width = w;
    height = h;
}

void Window3D::swapBuffers( void )
{
    imp->renderSurface().swapBuffers();

    imp->update();
}

void Window3D::sync( int n )
{
    imp->renderSurface().sync(n);
}

void Window3D::setKeyboardCallback( KeyboardCallback *callback )
{
    imp->setKeyboardCallback( callback );
}

int Window3D::mouseX(void)
{
    int  x, y;
    unsigned int w, h;
    imp->renderSurface().getWindowRectangle( x, y, w, h );
    return int((1 + imp->kbdMouseCallback().mx()) * float(w))>>1;
}

int Window3D::mouseY(void)
{
    int  x, y;
    unsigned int w, h;
    imp->renderSurface().getWindowRectangle( x, y, w, h );
    return int((1+imp->kbdMouseCallback().my()) * float(h))>>1;
}

unsigned int Window3D::mouseButton(void)
{
    return imp->kbdMouseCallback().mbutton();
}

void Window3D::enableTrackball( void )
{
    imp->enableTrackball();
}

void Window3D::disableTrackball( void )
{
    imp->disableTrackball();
}

float *Window3D::getTrackballMatrix(void)
{
    return imp->trackball()->getMatrix().ptr();
}

void Window3D::setTrackballScale( float s )
{
   imp->trackball()->setScale( s );
}


Trackball &Window3D::getTrackball( void ) 
{
    return *(imp->trackball());
};

bool Window3D::done() { return  imp->done(); }

